#!/usr/bin/env python
# -*- coding:utf-8 -*-


# ############################################################################
#  license :
# ============================================================================
#
#  File :        DonkiWeb.py
#
#  Project :     DonkiWeb: web to Tango
#
# This file is part of Tango device class.
# 
# Tango is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Tango is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with Tango.  If not, see <http://www.gnu.org/licenses/>.
# 
#
#  $Author :      sci.comp$
#
#  $Revision :    $
#
#  $Date :        $
#
#  $HeadUrl :     $
# ============================================================================
#            This file is generated by POGO
#     (Program Obviously used to Generate tango Object)
# ############################################################################

__all__ = ["DonkiWeb", "DonkiWebClass", "main"]

__docformat__ = 'restructuredtext'

import PyTango
import sys
# Add additional import
#----- PROTECTED REGION ID(DonkiWeb.additionnal_import) ENABLED START -----#
import os, time, signal
import traceback
from multiprocessing import Process
my_path = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, my_path+"/donkiweb_lib")
from donkiweb_srv import DonkiWebServer,DonkiWebsocketServer

class webServerThread (Process):
    def __init__(self, port_in, auth_file_in, web_base_path):
        super(webServerThread, self).__init__()
        self.server = DonkiWebServer(port_in, auth_file_in, web_base_path )

    def stop(self):
        self.server.stop()

    def run(self):
        self.server.start()

class websocketServerThread (Process):
    def __init__(self, port_in):
        super(websocketServerThread, self).__init__()
        self.server = DonkiWebsocketServer(port_in)

    def stop(self):
        self.server.stop()

    def run(self):
        self.server.start()

#----- PROTECTED REGION END -----#    //    DonkiWeb.additionnal_import

# Device States Description
# No states for this device


class DonkiWeb (PyTango.LatestDeviceImpl):
    """DonkiWeb: web to Tango"""
    
    # -------- Add you global variables here --------------------------
    #----- PROTECTED REGION ID(DonkiWeb.global_variables) ENABLED START -----#
    
    #----- PROTECTED REGION END -----#    //    DonkiWeb.global_variables

    def __init__(self, cl, name):
        PyTango.LatestDeviceImpl.__init__(self,cl,name)
        self.debug_stream("In __init__()")
        DonkiWeb.init_device(self)
        #----- PROTECTED REGION ID(DonkiWeb.__init__) ENABLED START -----#
        
        #----- PROTECTED REGION END -----#    //    DonkiWeb.__init__
        
    def delete_device(self):
        self.debug_stream("In delete_device()")
        #----- PROTECTED REGION ID(DonkiWeb.delete_device) ENABLED START -----#
        self.srvProc.stop()
        os.kill(self.srvProc_ws.pid, signal.SIGKILL)
        os.kill(self.srvProc.pid, signal.SIGKILL)
        
        #----- PROTECTED REGION END -----#    //    DonkiWeb.delete_device

    def init_device(self):
        self.debug_stream("In init_device()")
        self.get_device_properties(self.get_device_class())
        #----- PROTECTED REGION ID(DonkiWeb.init_device) ENABLED START -----#
        self.set_state(PyTango.DevState.ON)
        try:
            self.srvProc = webServerThread(self.Web_Port, self.Authentication_File, self.Web_Base_Path)
            self.srvProc.start()
            #
            if self.Websocket_Port < 0:
                return
            self.srvProc_ws = websocketServerThread(self.Websocket_Port)
            self.srvProc_ws.start()
            #
            self.set_state(PyTango.DevState.ON)
            self.debug_stream("In delete_device()")
        except:
            self.set_state(PyTango.DevState.FAULT)
            self.set_status(traceback.format_exc())
        
        #----- PROTECTED REGION END -----#    //    DonkiWeb.init_device

    def always_executed_hook(self):
        self.debug_stream("In always_excuted_hook()")
        #----- PROTECTED REGION ID(DonkiWeb.always_executed_hook) ENABLED START -----#
        
        #----- PROTECTED REGION END -----#    //    DonkiWeb.always_executed_hook

    # -------------------------------------------------------------------------
    #    DonkiWeb read/write attribute methods
    # -------------------------------------------------------------------------
    
    
    
            
    def read_attr_hardware(self, data):
        self.debug_stream("In read_attr_hardware()")
        #----- PROTECTED REGION ID(DonkiWeb.read_attr_hardware) ENABLED START -----#
        
        #----- PROTECTED REGION END -----#    //    DonkiWeb.read_attr_hardware


    # -------------------------------------------------------------------------
    #    DonkiWeb command methods
    # -------------------------------------------------------------------------
    

    #----- PROTECTED REGION ID(DonkiWeb.programmer_methods) ENABLED START -----#
    
    #----- PROTECTED REGION END -----#    //    DonkiWeb.programmer_methods

class DonkiWebClass(PyTango.DeviceClass):
    # -------- Add you global class variables here --------------------------
    #----- PROTECTED REGION ID(DonkiWeb.global_class_variables) ENABLED START -----#
    
    #----- PROTECTED REGION END -----#    //    DonkiWeb.global_class_variables


    #    Class Properties
    class_property_list = {
        }


    #    Device Properties
    device_property_list = {
        'Authentication_File':
            [PyTango.DevString, 
            "Configuration file for user and machine authentication",
            ["auth.config"] ],
        'Web_Port':
            [PyTango.DevShort, 
            "Web server port",
            [8080]],
        'Web_Base_Path':
            [PyTango.DevString, 
            "Reference  base path for the web content.",
            ["/var/www/easy_web_tango"] ],
        'Websocket_Port':
            [PyTango.DevShort, 
            "WebSocket server port (Negative Number to disable)",
            [8090]],
        }


    #    Command definitions
    cmd_list = {
        }


    #    Attribute definitions
    attr_list = {
        }


def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_class(DonkiWebClass, DonkiWeb, 'DonkiWeb')
        #----- PROTECTED REGION ID(DonkiWeb.add_classes) ENABLED START -----#
        
        #----- PROTECTED REGION END -----#    //    DonkiWeb.add_classes

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed:
        print ( '-------> Received a DevFailed exception: %s' % traceback.format_exc())
    except:
        print ('-------> An unforeseen exception occured.... %s' % traceback.format_exc())

if __name__ == '__main__':
    main()
