#!/usr/bin/env python3

import os,sys,time
import bottle
from bottle import route,run,template,response,request,HTTPError,static_file
from xml.dom import minidom
import PyTango
import traceback
from tangolistener import TangoDeviceListener
import socket
try:
    from urllib.request import urlopen
except:
    from urllib2 import urlopen
import traceback
import json
from websocket_server.websocket_server import WebsocketServer
import logging
from multiprocessing import Process
# Needed for JWT
import base64
import secrets, hmac, hashlib

tg_types_dict = PyTango.CmdArgType.values

html_address = {}
html_users = {}
html_auth_users = {}
cs_address = {}
cs_users = {}
cs_auth_users = {}
auth_users_history = {}
proxies = {}
attrs = {}
attr_values = {}
tango_dbs = {}

default_poll_interval = 2

update_password_interval = 600 # 10 minutes
last_pwd_update = 0

#############################################################################
####                             JWT section                             ####
#############################################################################
def base64url_decode(input):
    return base64.urlsafe_b64decode(input)

def base64url_encode(input):
    stringAsBytes = input.encode('ascii')
    stringAsBase64 = base64.urlsafe_b64encode(stringAsBytes).decode('utf-8').replace('=','')
    return stringAsBase64 

def jwt_creator(pwd, userid, userrole):
    header = {
        "alg": "HS256",
        "typ": "JWT"
    }
    payload = {'pwd': pwd,
               'userid': userid,
               'userrole': userrole,
               'expires': time.time() + 1800
               }
    secret_key = secrets.token_urlsafe(32)    
    total_params = str(base64url_encode(json.dumps(header))) + '.' + str(base64url_encode(json.dumps(payload)))
    signature = hmac.new(secret_key.encode(), total_params.encode(), hashlib.sha256).hexdigest()
    token = total_params + '.' + str(base64url_encode(signature))
    return token

def jwt_decoder(token):
    # Split by dot and get middle, payload, part;
    token_payload = token.split(".")[1]
    # Payload is base64 encoded, let's decode it to plain string
    # To make sure decoding will always work - we're adding max padding ("==")
    # to payload - it will be ignored if not needed.
    token_payload_decoded = str(base64.b64decode(token_payload + "=="), "utf-8")
    # Payload is JSON - we can load it to dict for easy access
    payload = json.loads(token_payload_decoded)
    return payload


#############################################################################
class DonkiWebsocketServer:
    def __init__(self, port):
        self._port = port
        self.server = WebsocketServer(host='0.0.0.0', port = self._port, loglevel=logging.NOTSET) #ERROR #CRITICAL #FATAL
        self.server.set_fn_new_client(self.new_client)
        self.server.set_fn_message_received(self.message_received)
        self.period_ms = 1000.
        self.attrs = []

    #############################################################################
    def start(self):
        self.server.run_forever(threaded=False)

    #############################################################################
    def stop(self):
        self.server.shutdown_abruptly()

    #############################################################################
    def new_client(self,client, server):
        server.send_message(client, json.dumps({"reply":"OK"}))

    #############################################################################
    def message_received(self,client, server, message):
        msg_tokens = message.split("&")
        request_query = {}
        for m in msg_tokens:
            request_query[m.split("=")[0]] = m.split("=")[1]
        if 'period' in request_query.keys():
            self.period_ms = float(request_query['period'])
        if 'td' in request_query.keys():
            self.attrs += (request_query['td']).split(",")
        if 'cmd' in request_query.keys():
            if (request_query['cmd'] == "start_streaming"):
                try:
                    self.tango_streaming(client, server, self.attrs, self.period_ms)
                except:
                    #traceback.print_exc()
                    pass

    #############################################################################
    def tango_streaming(self, client, server, attrs, period_ms):
        global proxies
        for fulltgattr in attrs:
            try:
                td,ta = fulltgattr.rsplit("/",1)
                if td.lower() not in proxies:
                    dev = TangoDeviceListener(td.lower(),default_poll_interval)
                    dev.start()
                    proxies[td.lower()] = dev
                if ta not in proxies[td.lower()].Attributes():
                    proxies[td.lower()].addAttribute(ta)
            except:
                pass
        prev_values = {}
        # It happens that the client page is not yet ready
        # added a small delay at startup
        # time.sleep(0.5)
        while True:
            out_dict = []
            for fulltgattr in attrs:
                try:
                    td,ta = fulltgattr.rsplit("/",1)
                    attr_reply = proxies[td.lower()].readAttribute(ta)
                    rdict = {}
                    if attr_reply is None:
                        rdict['value'] = None
                        rdict['name'] = fulltgattr
                    else:
                        rdict['name'] = fulltgattr
                        rdict['has_failed'] = attr_reply.has_failed
                        rdict['quality'] = str(attr_reply.quality)
                        rdict['type'] = int(attr_reply.type)
                        #rdict['type_name'] = str(tg_types_dict[int(attr_reply.type)])
                        rdict['r_dimension'] = (attr_reply.r_dimension.dim_x,attr_reply.r_dimension.dim_y)
                        rdict['unit'] = attr_reply.info.unit
                        if rdict['r_dimension'][0] > 1 or rdict['r_dimension'][1] > 0 :
                            if isinstance(attr_reply.value, tuple) or isinstance(attr_reply.value, list):
                                rdict['value'] = list(attr_reply.value)
                            else:
                                rdict['value'] = attr_reply.value.tolist()
                        else:
                            rdict['value'] = attr_reply.value
                        if (fulltgattr in prev_values.keys()):
                            prev_dict = prev_values[fulltgattr][0]
                            prev_time = prev_values[fulltgattr][1]
                            if (prev_dict == rdict) and (time.time() - prev_time) < 30:
                                # Skip notification for constant values (just time to time)
                                continue
                        prev_values[fulltgattr] = [rdict.copy(), time.time()]
                        rdict['time'] = attr_reply.time.totime()
                    out_dict.append(rdict)
                except:
                    #traceback.print_exc()
                    pass
            if len(out_dict):
                try:
                    reply = json.dumps(out_dict)
                except:
                    traceback.print_exc()
                    return
                try:
                    server.send_message(client, str(reply))
                except:
                    return
            time.sleep(period_ms/1000.)


#############################################################################
class DonkiWebServer:
    def __init__(self, port, auth_file, base_path):
        self._port = port
        self._auth_file = auth_file
        self._base_path = base_path
        self._app = bottle.Bottle()
        self._route()

    #############################################################################
    def _route(self):
        self._app.route('/', method="GET", callback=self._index)
        self._app.route('/tango-cs/<name>', callback=self._tango_rest)
        self._app.route('/html/<filepath:path>', callback=self._server_static)

    #############################################################################
    def stop(self):
        self._app.close()        

    #############################################################################
    def start(self):
        self.reload_config()
        self._app.run(host='0.0.0.0', port=self._port, quiet = True, server='auto')

    #############################################################################
    def _index(self):
        if os.path.isfile(self._base_path+'/html/index.html'):
            return static_file('index.html',root=self._base_path+'/html')
        else:
            return "Welcome! index.html not found in html folder"

    #############################################################################
    def _server_static(self,filepath):
        htmlpath = "/html/%s" % filepath
        reqpath = htmlpath.rsplit("/",1)[0]
        token_name = "token%d"%hash(reqpath)
        auth_cookie = request.get_cookie(token_name, secret="donkiweb_secret")
        if auth_cookie is not None:
            token_dict = jwt_decoder(auth_cookie)
            if (time.time() > token_dict['expires']):
                err = HTTPError(401, "Access denied")
                realm = "private"
                err.add_header('WWW-Authenticate', 'Basic realm="%s"' % realm)
                err.set_cookie(token_name, None, secret="donkiweb_secret",expires=time.time())
                return err
        auth_token = self.check_html_auth(htmlpath)
        if auth_token is None:
            err = HTTPError(401, "Access denied")
            realm = "private"
            err.add_header('WWW-Authenticate', 'Basic realm="%s"' % realm)
            err.set_cookie(token_name, None, secret="donkiweb_secret",expires=time.time())
            return err
        response = static_file(filepath,root=self._base_path+'/html')
        cookie_token = jwt_creator(auth_token[1], auth_token[0], auth_token[2])
        response.set_cookie(token_name, cookie_token, secret="donkiweb_secret",expires=time.time()+1800)
        return response

    #############################################################################
    def _tango_rest(self, name):
        global proxies, tango_dbs
        #
        if name.upper() in ['SETATTR','CMD']:
            token_name = "token%d"%hash(name)
            auth_cookie = request.get_cookie(token_name, secret="donkiweb_secret")
            if auth_cookie is not None:
                token_dict = jwt_decoder(auth_cookie)
                if (time.time() > token_dict['expires']):
                    err = HTTPError(401, "Access denied")
                    realm = "private"
                    err.add_header('WWW-Authenticate', 'Basic realm="%s"' % realm)
                    err.set_cookie(token_name, None, secret="donkiweb_secret",expires=time.time())
                    return err
            auth_token = self.check_cs_auth()
            if auth_token is None:
                err = HTTPError(401, "Access denied")
                realm = "private"
                err.add_header('WWW-Authenticate', 'Basic realm="%s"' % realm)
                err.set_cookie(token_name, None, secret="donkiweb_secret",expires=time.time())
                return err
            cookie_token = jwt_creator(auth_token[1], auth_token[0], auth_token[2])
            response.set_cookie(token_name, cookie_token, secret="donkiweb_secret",expires=time.time()+1800)
        #
        try:
            if name.upper() in ['GETATTR','GETFULLATTR']:
                out_dict = []
                for fulltgattr in (request.query['td']).split(","):
                    try:
                        td,ta = fulltgattr.rsplit("/",1)
                    except:
                        continue
                    if td.lower() not in proxies:
                        dev = TangoDeviceListener(td.lower(),default_poll_interval)
                        dev.start()
                        proxies[td.lower()] = dev
                    if ta not in proxies[td.lower()].Attributes():
                        proxies[td.lower()].addAttribute(ta)
                    if ('force' in request.query.keys()):
                        if request.query['force'].lower() == "true":
                            proxies[td.lower()].forceReadAttr(ta)
                    attr_reply = proxies[td.lower()].readAttribute(ta)
                    rdict = {}
                    try:
                        if attr_reply is None:
                            rdict['value'] = None
                            rdict['name'] = fulltgattr
                        elif name.upper() in ['GETFULLATTR']:
                            str_reply = str(attr_reply).rsplit("]",1)[0]
                            for i in str_reply.split("\n"):
                                if '=' in i:
                                    dkey = i.split("=",1)[0].strip()
                                    dval = i.split("=",1)[1].strip()
                                    rdict[dkey] = dval
                        else:
                            rdict['name'] = fulltgattr
                            v_dim = (attr_reply.r_dimension.dim_x+attr_reply.r_dimension.dim_y)
                            if (v_dim > 1):
                                if isinstance(attr_reply.value, tuple) or isinstance(attr_reply.value, list):
                                    rdict['value'] = list(attr_reply.value)
                                else:
                                    rdict['value'] = attr_reply.value.tolist()
                            else:
                                rdict['value'] = attr_reply.value
                            rdict['has_failed'] = attr_reply.has_failed
                            rdict['quality'] = str(attr_reply.quality)
                            rdict['type'] = int(attr_reply.type)
                            rdict['type_name'] = str(tg_types_dict[int(attr_reply.type)])
                            rdict['time'] = attr_reply.time.totime()
                            rdict['r_dimension'] = (attr_reply.r_dimension.dim_x,attr_reply.r_dimension.dim_y)
                            rdict['unit'] = attr_reply.info.unit
                        out_dict.append(rdict)
                    except:
                        traceback.print_exc()
                reply = json.dumps(out_dict)
            elif name.upper() == 'SETATTR':
                replydict = {}
                try:
                    argin_tokens = (request.query['td']).split(",", 1)
                    fulltgattr = argin_tokens[0]
                    value_str = argin_tokens[1]
                    attrname = (fulltgattr.split("/")[-1]).lower()
                    tangodev = (fulltgattr.rsplit("/",1)[0]).lower()
                    replydict['name'] = fulltgattr
                    if tangodev not in proxies.keys():
                        proxies[tangodev] = TangoDeviceListener(tangodev)
                        proxies[tangodev].start()
                    if attrname not in proxies[tangodev].Attributes():
                        proxies[tangodev].addAttribute(attrname)
                    dInfo = proxies[tangodev].readAttributeInfo(attrname)
                    wvalue = self._str_argin_to_value(value_str, dInfo)
                    proxies[tangodev].writeAttribute(attrname, wvalue)
                    # Force the background readout of the attribute 
                    proxies[tangodev].forceUpdateAttr(attrname)
                    replydict['time'] = time.time()
                    replydict['reply'] = "***OK"
                except:
                    replydict['reply'] = "***ERROR"
                    replydict['time'] = time.time()
                    replydict['info'] = traceback.format_exc()
                reply = json.dumps(replydict)
            elif name.upper() == 'ATTRINFO':
                replydict = {}
                try:
                    value_str = None
                    argin_tokens = (request.query['td']).split(",",1)
                    fulltgattr = argin_tokens[0]
                    attrname = (fulltgattr.split("/")[-1]).lower()
                    tangodev = (fulltgattr.rsplit("/",1)[0]).lower()
                    replydict['name'] = fulltgattr
                    if tangodev not in proxies.keys():
                        proxies[tangodev] = TangoDeviceListener(tangodev)
                        proxies[tangodev].start()
                    if attrname not in proxies[tangodev].Attributes():
                        proxies[tangodev].addAttribute(attrname)
                    retval = proxies[tangodev].readAttributeInfo(attrname)
                    if retval is not None:
                        replydict['data_type'] = int(retval.data_type)
                        replydict['data_type_name'] = str(tg_types_dict[int(retval.data_type)])
                        replydict['max_value'] = retval.max_value
                        replydict['min_value'] = retval.min_value
                    replydict['time'] = time.time()
                    replydict['reply'] = "***OK"
                except:
                    replydict['reply'] = "***ERROR"
                    replydict['time'] = time.time()
                    replydict['info'] = traceback.format_exc()
                reply = json.dumps(replydict)
            elif name.upper() == 'CMD':
                replydict = {}
                try:
                    value_str = None
                    argin_tokens =(request.query['td']).split(",",1)
                    fulltgname = argin_tokens[0]
                    if len(argin_tokens) > 1:
                        value_str = argin_tokens[1]
                    cmdname = (fulltgname.split("/")[-1]).lower()
                    tangodev = (fulltgname.rsplit("/",1)[0]).lower()
                    replydict['name'] = fulltgname
                    if tangodev not in proxies.keys():
                        proxies[tangodev] = TangoDeviceListener(tangodev)
                        proxies[tangodev].start()
                    if value_str is not None:
                        cInfo = proxies[tangodev].readCommandInfo(cmdname)
                        wvalue = self._str_argin_to_value(value_str, cInfo)
                        retval = proxies[tangodev].sendCommand(cmdname, wvalue)
                    else:
                        retval = proxies[tangodev].sendCommand(cmdname)
                    replydict['reply'] = "***OK"
                    replydict['time'] = time.time()
                    #
                    if isinstance(retval, tuple) or isinstance(retval, list):
                        replydict['value'] = list(retval)
                    elif hasattr(retval, 'tolist') and callable(retval.tolist):
                        replydict['value'] = retval.tolist()
                    elif "DevState" in str(type(retval)):
                        replydict['value'] = str(retval)
                    else:
                        replydict['value'] = retval
                except:
                    replydict['reply'] = "***ERROR"
                    replydict['time'] = time.time()
                    replydict['info'] = traceback.format_exc()
                reply = json.dumps(replydict)
            elif name.upper() == 'CMDINFO':
                replydict = {}
                try:
                    value_str = None
                    argin_tokens = (request.query['td']).split(",",1)
                    fulltgname = argin_tokens[0]
                    cmdname = (fulltgname.split("/")[-1]).lower()
                    tangodev = (fulltgname.rsplit("/",1)[0]).lower()
                    replydict['name'] = fulltgname
                    if tangodev not in proxies.keys():
                        proxies[tangodev] = TangoDeviceListener(tangodev)
                        proxies[tangodev].start()
                    retval = proxies[tangodev].readCommandInfo(cmdname)
                    if retval is not None:
                        replydict['in_type'] = int(retval.in_type)
                        replydict['in_type_name'] = str(tg_types_dict[int(retval.in_type)])
                        replydict['in_type_desc'] = retval.in_type_desc
                        replydict['out_type'] = int(retval.out_type)
                        replydict['out_type_name'] = str(tg_types_dict[int(retval.out_type)])
                        replydict['out_type_desc'] = retval.out_type_desc
                    replydict['reply'] = "***OK"
                except:
                    replydict['reply'] = "***ERROR"
                    replydict['info'] = traceback.format_exc()
                reply = json.dumps(replydict)
            elif name.upper() == 'GETFULLINFO':
                replydict = {}
                try:
                    replydict['reply'] = "***OK"
                    value_str = None
                    argin_tokens = (request.query['td']).split(",",1)
                    fulltgname = argin_tokens[0]
                    tangodev = fulltgname.lower()
                    replydict['name'] = tangodev
                    if tangodev not in proxies.keys():
                        proxies[tangodev] = TangoDeviceListener(tangodev)
                        proxies[tangodev].start()
                    retval = proxies[tangodev].readDeviceFullInfo()
                    if retval is not None:
                        attrsInfo = {}
                        for attr in retval['Attributes']:
                            attrsInfo[attr.name] = {}
                            attrsInfo[attr.name]['data_format'] = int(attr.data_format)
                            attrsInfo[attr.name]['data_type'] = int(attr.data_type)
                            attrsInfo[attr.name]['data_type_name'] = str(tg_types_dict[int(attr.data_type)])
                            attrsInfo[attr.name]['max_value'] = attr.max_value
                            attrsInfo[attr.name]['min_value'] = attr.min_value
                            attrsInfo[attr.name]['description'] = attr.description
                            attrsInfo[attr.name]['format'] = attr.format
                            attrsInfo[attr.name]['display_unit'] = attr.display_unit
                            attrsInfo[attr.name]['standard_unit'] = attr.standard_unit
                            attrsInfo[attr.name]['unit'] = attr.unit
                            attrsInfo[attr.name]['writable'] = int(attr.writable)
                            attrsInfo[attr.name]['max_dim_x'] = int(attr.max_dim_x)
                            attrsInfo[attr.name]['max_dim_y'] = int(attr.max_dim_y)
                        replydict['Attributes'] = attrsInfo
                        #
                        cmdsInfo = {}
                        for cmd in retval['Commands']:
                            cmdsInfo[cmd.cmd_name] = {}
                            cmdsInfo[cmd.cmd_name]['in_type'] = int(cmd.in_type)
                            cmdsInfo[cmd.cmd_name]['in_type_name'] = str(tg_types_dict[int(cmd.in_type)])
                            cmdsInfo[cmd.cmd_name]['in_type_desc'] = cmd.in_type_desc
                            cmdsInfo[cmd.cmd_name]['out_type'] = int(cmd.out_type)
                            cmdsInfo[cmd.cmd_name]['out_type_name'] = str(tg_types_dict[int(cmd.out_type)])
                            cmdsInfo[cmd.cmd_name]['out_type_desc'] = cmd.out_type_desc
                        replydict['Commands'] = cmdsInfo
                        #
                        replydict['info'] = retval['Info']
                        #
                        if retval['Error'] != "":
                            replydict['reply'] = "***ERROR"
                except:
                    replydict['reply'] = "***ERROR"
                    replydict['info'] = traceback.format_exc()
                reply = json.dumps(replydict)
            elif name.upper() == 'STREAM':
                response.content_type = "text/event-stream"
                response.cache_control = 'no-cache'
                period_ms = 1000.
                if 'period' in request.query.keys():
                    period_ms = float(request.query['period'])
                attrs = (request.query['th']).split(",")
                return self.tango_streaming(attrs,period_ms)
            elif name.upper() == 'BROWSE':
                query = request.query
                # Check input parameter
                if "th" not in query.keys():
                    tg_host = "notdefined"
                    if "notdefined" not in tango_dbs.keys():
                        tango_dbs["notdefined"] = PyTango.Database()
                else:
                    TH = query["th"].split("/")[0]
                    tg_h = TH.split(":")[0]
                    tg_p = TH.split(":")[1]
                    tg_host = "%s:%s" % (tg_h,tg_p)
                    if tg_host not in tango_dbs.keys():
                        tango_dbs[tg_host] = PyTango.Database(tg_h,tg_p)
                #
                full_info = False
                if "fullinfo" in query.keys():
                    full_info = (query["fullinfo"].lower() in ["true","1"])
                #
                replydict = {}
                tds = []
                for srv_name in tango_dbs[tg_host].get_server_list("*"):
                    devs_list = (tango_dbs[tg_host].get_device_class_list(srv_name)).value_string
                    for i in range(int(len(devs_list)/2)):
                        if full_info:
                            tds.append([devs_list[i*2],devs_list[i*2+1],srv_name])
                        else:
                            tds.append(devs_list[i*2])
                tds.sort()
                replydict['value'] = str(tds)
                replydict['reply'] = "***OK"
                reply = json.dumps(replydict)
            else:
                replydict = {}
                replydict['reply'] = '?'
                reply = json.dumps(replydict)
        except:
            replydict = {}
            replydict['reply'] = "***ERROR"
            replydict['info'] = traceback.format_exc()
            reply = json.dumps(replydict)
        response.content_type = 'text/plain'
        return str(reply)


    #############################################################################
    def tango_streaming(self,attrs,period_ms):
        global proxies
        for fulltgattr in attrs:
            td,ta = fulltgattr.rsplit("/",1)
            if td.lower() not in proxies:
                dev = TangoDeviceListener(td.lower(),default_poll_interval)
                dev.start()
                proxies[td.lower()] = dev
            if ta not in proxies[td.lower()].Attributes():
                proxies[td.lower()].addAttribute(ta)
        # Set client-side auto-reconnect timeout, ms.
        #yield 'retry: 100\n\n'
        prev_values = {}
        while True:
            out_dict = []
            for fulltgattr in attrs:
                td,ta = fulltgattr.rsplit("/",1)
                attr_reply = proxies[td.lower()].readAttribute(ta)
                rdict = {}
                try:
                    if attr_reply is None:
                        rdict['value'] = None
                        rdict['name'] = fulltgattr
                    else:
                        rdict['name'] = fulltgattr
                        rdict['value'] = attr_reply.value
                        rdict['has_failed'] = attr_reply.has_failed
                        rdict['quality'] = str(attr_reply.quality)
                        rdict['type'] = int(attr_reply.type)
                        rdict['r_dimension'] = (attr_reply.r_dimension.dim_x,attr_reply.r_dimension.dim_y)
                        rdict['unit'] = attr_reply.info.unit
                        if (fulltgattr in prev_values.keys()):
                            prev_dict = prev_values[fulltgattr][0]
                            prev_time = prev_values[fulltgattr][1]
                            if (prev_dict == rdict) and (time.time() - prev_time) < 30:
                                # Skip notification for constant values (just time to time)
                                continue
                        prev_values[fulltgattr] = [rdict.copy(), time.time()]
                        rdict['time'] = attr_reply.time.totime()
                    out_dict.append(rdict)
                except:
                    traceback.print_exc()
            if len(out_dict):
                reply = json.dumps(out_dict)
                yield 'data: %s\n\n' % str(reply)
            time.sleep(period_ms/1000.)

    #############################################################################
    def _str_argin_to_value(self, argin, dInfo):
        BOOL_TYPES = [1]
        INT_TYPES = [2,3,6,7,22,23,24,27]
        FLOAT_TYPES = [4,5]
        STRING_TYPES = [8]
        SPECTRUM_TYPES = [9,10,11,12,13,14,15,25,26];
        SPECTRUM_STRING_TYPES = [16];
        STATE_TYPES = [19]

        if hasattr(dInfo, 'data_type'):
            # attribute
            dtype = dInfo.data_type
            dim_x = dInfo.max_dim_x
            dim_y = dInfo.max_dim_y
        else:
            # command
            dtype = dInfo.in_type
            dim_x = 1
            dim_y = 0
        if (dtype in BOOL_TYPES):
            argin = argin.lower().replace('true','1')
            argin = argin.lower().replace('false','0')
            return eval(argin)
        elif dtype in INT_TYPES:
            return eval(argin)
        elif dtype in FLOAT_TYPES:
            return eval(argin)
        elif dtype in STRING_TYPES and (dim_x + dim_y) == 1:
            return str(argin)
        return eval(argin)


    #############################################################################
    def reload_config(self):
        global html_address,html_users,html_auth_users,last_pwd_update
        global cs_address,cs_users,cs_auth_users
        global last_pwd_update
        last_pwd_update = time.time()
        self._app.config.load_config(self._auth_file)
        for k in self._app.config.keys():
            if k.startswith('html_address.'):
                if self._app.config[k] is not None:
                    folder_info = k.split('html_address.',1)[-1]
                    html_address[folder_info] = []
                    for host in (self._app.config[k]).split(","):
                        try:
                            html_address[folder_info].append(socket.gethostbyname(host.strip()))
                        except:
                            html_address[folder_info].append(host.strip())
            elif k.startswith('html_user.'):
                folder_info = k.split('html_user.',1)[-1]
                html_users[folder_info] = []
                if self._app.config[k] is not None:
                    html_users[folder_info] = self._app.config[k].split(",")
            elif k.startswith('html_auth_user.'):
                folder_info = k.split('html_auth_user.',1)[-1]
                html_auth_users[folder_info] = []
                if self._app.config[k] is not None:
                    html_auth_users[folder_info] = self._app.config[k].split(",")
            elif k.startswith('cs_address.'):
                if self._app.config[k] is not None:
                    try:
                        tghost_info = socket.gethostbyname(k.split('cs_address.',1)[-1])
                    except:
                        tghost_info = (k.split('cs_address.',1)[-1]).strip()
                    cs_address[tghost_info] = []
                    for host in (self._app.config[k]).split(","):
                        try:
                            cs_address[tghost_info].append(socket.gethostbyname(host.strip()))
                        except:
                            cs_address[tghost_info].append(host.strip())
            elif k.startswith('cs_user.'):
                try:
                    tghost_info = socket.gethostbyname(k.split('cs_user.',1)[-1])
                except:
                    tghost_info = (k.split('cs_user.',1)[-1]).strip()
                cs_users[tghost_info] = []
                if self._app.config[k] is not None:
                    cs_users[tghost_info] = self._app.config[k].split(",")
            elif k.startswith('cs_auth_user.'):
                try:
                    tghost_info = socket.gethostbyname(k.split('cs_auth_user.',1)[-1])
                except:
                    tghost_info = (k.split('cs_auth_user.',1)[-1]).strip()
                cs_auth_users[tghost_info] = []
                if self._app.config[k] is not None:
                    cs_auth_users[tghost_info] = self._app.config[k].split(",")

    #############################################################################
    def check_cs_auth(self):
        global cs_address,cs_users,cs_auth_users
        global last_pwd_update, auth_users_history
        userrole = "user"
        if self._auth_file is None:
            return ['','',userrole]
        # Let's refresh passwords time.time()
        if (time.time() - last_pwd_update) > update_password_interval:
            self.reload_config()
        #
        argin_tokens = (request.query['td']).split(",",1)
        fulltgname = (request.query['td']).split(",",1)[0]
        tangodev = (fulltgname.rsplit("/",1)[0]).lower()
        tg_host = None
        if (len(tangodev.split("/")) > 3):
            tg_host = tangodev.split("/")[0]
        elif "TANGO_HOST" in os.environ.keys():
            tg_host = os.environ["TANGO_HOST"]
        try:
            tg_host_addr = socket.gethostbyname(tg_host.split(":")[0])
        except:
            tg_host_addr = tg_host.split(":")[0]
        # Check if the TANGO HOST is protected
        if (tg_host_addr not in cs_address.keys()) \
           and (tg_host_addr not in cs_users.keys()) \
           and (tg_host_addr not in cs_auth_users.keys()):
            return None
        # Check allowed IP addresses
        client_ip = socket.gethostbyname(request.environ.get('REMOTE_ADDR'))
        if (tg_host_addr in cs_address.keys()):
            if "*.*.*.*" in (cs_address[tg_host_addr]):
                return ['','',userrole]
            if client_ip in (cs_address[tg_host_addr]):
                return ['','',userrole]
        # Check authorized users
        user, pw = request.auth or (None, None)
        if (tg_host_addr in cs_users.keys()):
            for auth_token in cs_users[tg_host_addr]:
                try:
                    auth_info = auth_token.replace("(","").replace(")","").split("/",1)
                    if (auth_info[0] == user) and (auth_info[1] == pw):
                        return [user,pw,userrole]
                except:
                    traceback.print_exc()
        # Check authorized users
        if (tg_host_addr in cs_auth_users.keys()):
            if user in cs_auth_users[tg_host_addr]:
                # Let's try the Auth Service Login
                if user not in auth_users_history.keys():
                    try:
                        params = "FRM_USERNAME=%s&FRM_PASSWORD=%s" % (user,pw)
                        full_url = "%s?%s" % (self._app.config["auth_service.vuourl"], params)
                        url = urlopen(full_url)
                        argout = url.read()
                        if 'OK' in argout.decode("utf-8"):
                            # Clear time to time the cache?
                            auth_users_history[user] = pw
                        else:
                            return None
                    except:
                        traceback.print_exc()
                        return None
                if (pw == auth_users_history[user]):
                    return [user,pw,userrole]
        return None


    #############################################################################
    def check_html_auth(self, html_path):
        global html_address,html_users,html_auth_users,last_pwd_update
        global last_pwd_update, auth_users_history
        userrole = "user"
        if self._auth_file is None:
            return ['','',userrole]
        # Let's refresh passwords time.time()
        if (time.time() - last_pwd_update) > update_password_interval:
            self.reload_config()
        #
        reqpath = html_path.rsplit("/",1)[0]
        # Check if the HOST is protected
        if (reqpath not in html_address.keys()) \
           and (reqpath not in html_users.keys()) \
           and (reqpath not in html_auth_users.keys()):
            return ['','',userrole]
        # Check allowed IP addresses
        user, pw = request.auth or (None, None)
        client_ip = socket.gethostbyname(request.environ.get('REMOTE_ADDR'))
        client_user_agent = request.environ.get('HTTP_USER_AGENT')
        if (reqpath in html_address.keys()):
            if client_ip in (html_address[reqpath]):
                return ['','',userrole]
        # Check authorized users
        if (reqpath in html_users.keys()):
            for auth_token in html_users[reqpath]:
                try:
                    auth_info = auth_token.replace("(","").replace(")","").split("/",1)
                    if (auth_info[0] == user) and (auth_info[1] == pw):
                        return [user,pw,userrole]
                except:
                    traceback.print_exc()
        # Check authorized users
        if (reqpath in html_auth_users.keys()):
            if user in html_auth_users[reqpath]:
                # Let's try the Auth Service Login
                if user not in auth_users_history.keys():
                    try:
                        params = "FRM_USERNAME=%s&FRM_PASSWORD=%s" % (user,pw)
                        full_url = "%s?%s" % (self._app.config["auth_service.vuourl"], params)
                        url = urlopen(full_url)
                        argout = url.read()
                        if 'OK' in argout.decode("utf-8"):
                            # Clear time to time the cache?
                            auth_users_history[user] = [pw, client_ip, client_user_agent, time.time()]
                        else:
                            return None
                    except:
                        traceback.print_exc()
                        return None
                if (pw == auth_users_history[user][0]):
                    return [user,pw,userrole]
        return None

########################################################################
###     Code for console usage                                       ###
########################################################################

class webServerThread (Process):
    def __init__(self, port_in, auth_file_in, web_base_path):
        super(webServerThread, self).__init__()
        self.server = DonkiWebServer(port_in, auth_file_in, web_base_path )
        print("Web server listening on port %d" % port_in)
        print("Web base path %s" % web_base_path)
        print("Auth config file %s" % auth_file_in)

    def stop(self):
        self.server.stop()

    def run(self):
        self.server.start()

class websocketServerThread (Process):
    def __init__(self, port_in):
        super(websocketServerThread, self).__init__()
        self.server = DonkiWebsocketServer(port_in)
        print("WebSocket server listening on port %d" % port_in)

    def stop(self):
        self.server.stop()

    def run(self):
        self.server.start()


class DonkiWeb ():

    def __init__(self, Web_Port, Websocket_Port, Authentication_File, Web_Base_Path):
        print("Starting servers()")
        try:
            self.srvProc = webServerThread(Web_Port, Authentication_File, Web_Base_Path)
            self.srvProc.start()
            #
            if Websocket_Port < 0:
                return
            self.srvProc_ws = websocketServerThread(Websocket_Port)
            self.srvProc_ws.start()
        except:
            print (traceback.format_exc())
        
    def delete_device(self):
        self.srvProc.stop()
        os.kill(self.srvProc_ws.pid, signal.SIGKILL)
        os.kill(self.srvProc.pid, signal.SIGKILL)


def main():
    import argparse
    #
    web_port = 8080
    websocket_port = 8090
    auth_file = "auth.config"
    web_base_path = "/var/www/donkiweb"
    #
    parser = argparse.ArgumentParser()
    parser.add_argument('--web_port', type=int, help="Web server port")
    parser.add_argument('--websocket_port', type=int, help="WebSocket server port (Negative Number to disable)")
    parser.add_argument('--auth_file', type=str, help="Configuration file for user and machine authentication")
    parser.add_argument('--web_base_path', type=str, help="Reference  base path for the web content.")
    args = parser.parse_args()
    if args.web_port:
        web_port = args.web_port
    if args.websocket_port:
        websocket_port = args.websocket_port
    if args.auth_file:
        auth_file = args.auth_file
    if args.web_base_path:
        web_base_path = args.web_base_path
    dw = DonkiWeb(web_port, websocket_port, auth_file, web_base_path)

if __name__ == '__main__':
    main()
